from ctypes import cdll
from ctypes import create_string_buffer
from typing import Any

from loguru import logger

libc = cdll.LoadLibrary("libc.so.6")
MAX_QUEUE_LEN = 8192


class MessageQueue:
    """A simple abstraction over mq_* libc functions."""

    def __init__(self, rx: str, tx: str) -> None:
        self._rx = rx
        self._tx = tx
        self._rx_queue: int | None = None
        self._tx_queue: int | None = None

    def __enter__(self) -> "MessageQueue":
        self._rx_queue = libc.mq_open(bytes(self._rx, "utf-8"), 0, 666, None)
        if not self._is_rx_valid():
            libc.perror(b"mq_open")
            logger.error("bad rx open")
        self._tx_queue = libc.mq_open(bytes(self._tx, "utf-8"), 1, 666, None)
        if not self._is_tx_valid():
            libc.perror(b"mq_open")
            logger.error("bad tx open")
        return self

    def __exit__(self, *_: list[Any]) -> None:
        if self._is_rx_valid():
            libc.mq_close(self._rx_queue)
        if self._is_tx_valid():
            libc.mq_close(self._tx_queue)

    def send(self, msg: str) -> None:
        if not self._is_tx_valid():
            raise RuntimeError(f"tx queue is invalid {self._tx_queue}")
        res = libc.mq_send(self._tx_queue, bytes(msg, "utf-8"), len(msg), 1)
        if res == -1:
            libc.perror(b"mq_send")
            logger.error("bad send")

    def receive(self) -> str | None:
        if not self._is_rx_valid():
            raise RuntimeError(f"rx queue is invalid {self._rx_queue}")
        result = create_string_buffer(b"\0" * MAX_QUEUE_LEN)
        res = libc.mq_receive(self._rx_queue, result, MAX_QUEUE_LEN, None)
        if res == -1:
            libc.perror(b"mq_receive")
            logger.error("bad receive")
            return None
        return str(result.value, encoding="utf-8")

    def _is_rx_valid(self) -> bool:
        return self._rx_queue is not None and self._rx_queue != -1

    def _is_tx_valid(self) -> bool:
        return self._tx_queue is not None and self._tx_queue != -1
