import asyncio
from collections.abc import Callable
from typing import NamedTuple

from loguru import logger

from scii.mq import MessageQueue


class PipelineDTO(NamedTuple):
    id: str
    start_time: str


class _PipelinesDiff(NamedTuple):
    additions: dict[str, PipelineDTO]
    deletions: dict[str, PipelineDTO]


class SciPipelines:
    """SCI pipelines synchronizer class. Use this to automatically stay up to date."""

    def __init__(
        self,
        pipeline_new_callback: Callable[[PipelineDTO], None],
        pipeline_end_callback: Callable[[str, int | None], None],
    ) -> None:
        self._on_new_pipeline = pipeline_new_callback
        self._on_end_pipeline = pipeline_end_callback
        self._pipelines: dict[str, PipelineDTO] = {}
        self._queue = MessageQueue(rx="/sci_tx", tx="/sci_rx")
        self._is_running: bool = False

    async def async_run(self) -> None:
        await asyncio.to_thread(self.run)

    async def async_request_list(self) -> None:
        await asyncio.to_thread(self.request_list)

    def request_list(self) -> None:
        if self._is_running:
            self._queue.send("list")

    def run(self) -> None:
        with self._queue:
            try:
                self._is_running = True
                while True:
                    msg = self._queue.receive()
                    if msg is None:
                        logger.warning("did not receive anything")
                        continue
                    self._handle_message(msg)
            finally:
                self._is_running = False

    def _handle_message(self, msg: str) -> None:
        msg_type = msg.split(" ")[0]
        content = msg[len(msg_type) :]
        logger.trace("handling sci message {}", msg_type)
        match msg_type:
            case "list":
                pipelines_raw = content.split("\\n")
                pipelines = self._from_raw(pipelines_raw)
                diff = self._diff(self._pipelines, pipelines)
                for key in diff.additions:
                    self._on_new_pipeline(diff.additions[key])
                for key in diff.deletions:
                    self._on_end_pipeline(key, None)
                self._pipelines = pipelines

            case "pipeline_new":
                pipeline = self._from_raw2(content)
                if pipeline is not None:
                    self._on_new_pipeline(pipeline)

            case "pipeline_end":
                raw = content.split(" ")
                id = raw[0]
                return_code = None
                try:
                    return_code = int(raw[2])
                except ValueError:
                    logger.error("bad cast")
                self._on_end_pipeline(id, return_code)

            case "sci_started":
                for k in self._pipelines:
                    self._on_end_pipeline(k, None)
                self._pipelines.clear()

            case _:
                logger.warning("unknown sci message {}", msg)

    def _from_raw(self, raw: list[str]) -> dict[str, PipelineDTO]:
        result: dict[str, PipelineDTO] = {}
        for x in raw:
            if x == "":
                continue
            dto = self._from_raw2(x)
            if dto is not None:
                result[dto.id] = dto
        return result

    def _from_raw2(self, raw: str) -> PipelineDTO | None:
        if raw == "":
            return None
        s = raw.split(" ")
        return PipelineDTO(id=s[0], start_time=s[1])

    def _diff(self, a: dict[str, PipelineDTO], b: dict[str, PipelineDTO]) -> _PipelinesDiff:
        result = _PipelinesDiff(additions={}, deletions={})
        for akey in a:
            if akey not in b:
                result.deletions[akey] = a[akey]
        for bkey in b:
            if bkey not in a:
                result.additions[bkey] = b[bkey]
        return result
