from __future__ import annotations

import asyncio
from typing import override

import loguru
from loguru import logger
from textual.app import App
from textual.app import ComposeResult
from textual.containers import ScrollableContainer
from textual.driver import Driver
from textual.types import CSSPathType
from textual.widgets import Footer
from textual.widgets import Header
from textual.widgets import Log

from scii.sci_pipelines import PipelineDTO
from scii.sci_pipelines import SciPipelines
from scii.tui.widgets.pipeline import PipelineWidget


class _WritableLog(loguru.Writable):
    def __init__(self, log: Log) -> None:
        super().__init__()
        self._log = log

    @override
    def write(self, message: loguru.Message) -> None:
        _ = self._log.write(message)


class TuiApp(App[None]):
    """A Textual app to manage sci pipelines."""

    BINDINGS = [
        ("d", "toggle_dark", "Toggle dark mode"),
        ("l", "toggle_log", "Toggle log"),
        ("q", "quit", "Quit"),
    ]

    def __init__(
        self,
        driver_class: type[Driver] | None = None,
        css_path: CSSPathType | None = None,
        watch_css: bool = False,
        ansi_color: bool = False,
    ):
        super().__init__(driver_class, css_path, watch_css, ansi_color)
        self._pipelines = SciPipelines(self._on_new_pipeline, self._on_end_pipeline)
        self._my_log: Log = Log(id="log")
        self._my_log_wrapper = _WritableLog(self._my_log)

    def setup_logging(self, level: str, use_colors: bool) -> "TuiApp":
        """Setup logging.

        Args:
            level: The logging level.
            use_colors: Whether to use colors in the log output.

        Returns:
            self

        """
        logger.remove()
        _ = logger.add(
            self._my_log_wrapper,
            colorize=use_colors,
            format="{time:HH:mm:ss} <level>{level}</level> <fg #888>{file}:{line}:</fg #888> <bold>{message}</bold>",
            level=level
        )
        return self

    def on_mount(self) -> None:
        _ = asyncio.create_task(self._pipelines.async_run())  # run the pipelines task in the background
        _ = self.set_interval(1, self._pipelines.async_request_list)

    @override
    def compose(self) -> ComposeResult:
        """Create child widgets for the app."""
        yield Header()
        yield Footer()
        yield ScrollableContainer(id="pipelines")
        yield self._my_log

    @override
    def action_toggle_dark(self) -> None:
        """An action to toggle dark mode."""
        self.dark = not self.dark

    def action_toggle_log(self) -> None:
        self._my_log.display = not self._my_log.display

    async def action_add_pipeline(self, id: str, start_time: str | None = None) -> None:
        """An action to add a timer."""
        if start_time is None:
            start_time = "N/A"
        new_pipeline = PipelineWidget(id, start_time)
        _ = self.query_one("#pipelines").mount(new_pipeline)

    async def action_remove_pipeline(self, id: str) -> None:
        """Called to remove a timer."""
        containers = self.query_one("#pipelines")
        for child in containers.children:
            if not isinstance(child, PipelineWidget):
                continue
            if child.pipeline_id == id:
                logger.trace("removing pipeline widget with id {}", id)
                await child.remove()

    def _on_new_pipeline(self, pipeline: PipelineDTO) -> None:
        logger.debug("TODO: add pipeline {} to tui", pipeline)

    def _on_end_pipeline(self, id: str, return_code: int | None) -> None:
        logger.debug("TODO: remove pipeline {}({}) from tui", id, return_code)
